/***************************************************************************
                          cntfspart.h  -  description
                             -------------------
    begin                : Mon May 22 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CNTFSPART_H
#define CNTFSPART_H

#include "partimage.h"
#include "common.h"
#include "misc.h"

#include <asm/types.h>

struct CImage;
struct CVolumeHeader;
struct CMainFooter;
struct CMainHeader;
struct COptions;
class CSaveWindow;

#define CURRENT_NTFS_FORMAT            "0.1" // format version of the ntfs image
#define MAX_NTFS_BLOCK_SIZE 8192

#ifdef __LITTLE_ENDIAN

#define CPU_TO_LE16(a) (a)
#define CPU_TO_LE32(a) (a)
#define CPU_TO_LE64(a) (a)
#define LE16_TO_CPU(a) (a)
#define LE32_TO_CPU(a) (a)
#define LE64_TO_CPU(a) (a)

#else*/

/* Routines for big-endian machines */
#ifdef __BIG_ENDIAN

/* We hope its big-endian, not PDP-endian :) */
#define CPU_TO_LE16(a) ((((a)&0xFF) << 8)|((a) >> 8))
#define CPU_TO_LE32(a) ((((a) & 0xFF) << 24) | (((a) & 0xFF00) << 8) | \
			(((a) & 0xFF0000) >> 8) | ((a) >> 24))
#define CPU_TO_LE64(a) ((CPU_TO_LE32(a)<<32)|CPU_TO_LE32((a)>>32)

#define LE16_TO_CPU(a) CPU_TO_LE16(a)
#define LE32_TO_CPU(a) CPU_TO_LE32(a)
#define LE64_TO_CPU(a) CPU_TO_LE64(a)

#endif
#endif

typedef BYTE  ntfs_u8;
typedef UINT ntfs_u16;
typedef DWORD ntfs_u32;
typedef QWORD ntfs_u64;
typedef signed char ntfs_s8;
typedef signed short ntfs_s16;
typedef signed long ntfs_s32;
typedef signed long long ntfs_s64;

#define NTFS_GETU8(p)      (*(ntfs_u8*)(p))
#define NTFS_GETU16(p)     ((ntfs_u16)LE16_TO_CPU(*(ntfs_u16*)(p)))
#define NTFS_GETU24(p)     ((ntfs_u32)NTFS_GETU16(p) | ((ntfs_u32)NTFS_GETU8(((char*)(p))+2)<<16))
#define NTFS_GETU32(p)     ((ntfs_u32)LE32_TO_CPU(*(ntfs_u32*)(p)))
#define NTFS_GETU40(p)     ((ntfs_u64)NTFS_GETU32(p)|(((ntfs_u64)NTFS_GETU8(((char*)(p))+4))<<32))
#define NTFS_GETU48(p)     ((ntfs_u64)NTFS_GETU32(p)|(((ntfs_u64)NTFS_GETU16(((char*)(p))+4))<<32))
#define NTFS_GETU56(p)     ((ntfs_u64)NTFS_GETU32(p)|(((ntfs_u64)NTFS_GETU24(((char*)(p))+4))<<32))
#define NTFS_GETU64(p)     ((ntfs_u64)LE64_TO_CPU(*(ntfs_u64*)(p)))


#define NTFS_GETS8(p)        ((*(ntfs_s8*)(p)))
#define NTFS_GETS16(p)       ((ntfs_s16)LE16_TO_CPU(*(short*)(p)))
#define NTFS_GETS24(p)       (NTFS_GETU24(p) < 0x800000 ? (int)NTFS_GETU24(p) : (int)(NTFS_GETU24(p) - 0x1000000))
#define NTFS_GETS32(p)       ((ntfs_s32)LE32_TO_CPU(*(int*)(p)))
#define NTFS_GETS40(p)       (((ntfs_s64)NTFS_GETU32(p)) | (((ntfs_s64)NTFS_GETS8(((char*)(p))+4)) << 32))
#define NTFS_GETS48(p)       (((ntfs_s64)NTFS_GETU32(p)) | (((ntfs_s64)NTFS_GETS16(((char*)(p))+4)) << 32))
#define NTFS_GETS56(p)       (((ntfs_s64)NTFS_GETU32(p)) | (((ntfs_s64)NTFS_GETS24(((char*)(p))+4)) << 32))
#define NTFS_GETS64(p)	     ((ntfs_s64)NTFS_GETU64(p))

// ================================================
class CNtfsRunList
{
private:
	QWORD *m_qwOffset;
	QWORD m_qwClustersCount;
	
public:
	CNtfsRunList();
	~CNtfsRunList();
	
public:
	int setCluster(QWORD qwNumber, QWORD qwOffset); // ex: cluser 0 position is 12589
	int init(QWORD qwClustersCount);
	void freeMemory();
	void show();
	QWORD offset(QWORD qwCluster);
	QWORD clustersCount();
};

// ================================================
class CNtfsPart
{
public:
	CNtfsPart(char *szDevice, FILE *fDeviceFile);
	~CNtfsPart();
	int printfInformations();
	int readBootSector();
	int readBitmap();
		
	int saveUsedBlocksToImageFile(COptions *options);
	int restoreUsedBlocksToDisk(COptions *options);

	// save options
	int saveToImage(CImage *image, COptions *options, CSaveWindow *gui);

	// restore options
	int restoreFromImage(CImage *image, COptions *options, CRestoreWindow *gui);

	// file reading options
	int readDataAttribute(BYTE *cAttribData, CNtfsRunList *runlist, QWORD *qwBitmapSize, QWORD *qwAllocatedSize);
	int readFilenameAttribute(BYTE *cAttribData, char *szFilename);
	int readFileRecord(BYTE *cRecordData, CNtfsRunList *runlist, QWORD *qwBitmapSize, QWORD *qwAllocatedSize);
	int processRunList(BYTE *cAttribData, CNtfsRunList *runlist);
	int searchForRecordFilename(BYTE *cFileRecord, char *szFilename, char *szUnicodeName, int nUnicodeNameLen);
	int checkFilenameForRecordIs(BYTE *cRecordData, char *szFilename, bool bShowDebug = true);
	int readVolumeLabel();

private:
	CImage *m_image;
	CBlocksUseBitmap m_bitmap;
	CSaveWindow *m_guiSave;
	CRestoreWindow *m_guiRestore;
	
	char m_szDevice[MAX_DEVICENAMELEN+1]; // "/dev/hdc5"
	char m_szImageFilename[MAXPATHLEN+1]; // ex: /home/user/filename.partimg.gz.002 (if splitted)
	char m_szOriginalFilename[MAXPATHLEN+1]; // ex: filename.partimg.gz
	char m_szPath[MAXPATHLEN+1]; // ex: /home/user/
	
	// informations
	UINT m_nBytesPerSector;
	BYTE m_cSectorsPerCluster;
	QWORD m_qwTotalSectorsCount;
	QWORD m_qwLCNOfMftDataAttrib;
	DWORD m_dwFileRecordSize;
	
	QWORD m_qwTotalClustersCount;
	QWORD m_qwUsedClustersCount;
	QWORD m_qwFreeClustersCount;
	
	DWORD m_dwBitmapSize;
	
	char m_szLabel[128];
	UINT m_nNtfsVersion;
			

	DWORD m_dwClusterSize;
	
	// volumes
	DWORD m_dwCurrentVolumeSize;
	DWORD m_dwVolumeNumber;
		
	// File descriptors
	FILE *m_fDeviceFile;
};

// ================================================
struct CNtfsHeader // size must be 16384 (adjust the reserved data)
{
	char szAppFormat[20]; // Version of the format of the application file
	DWORD dwNtfsFlags;
	
	UINT nBytesPerSector;
	BYTE cSectorsPerCluster;
	QWORD qwTotalSectorsCount;
	QWORD qwLCNOfMftDataAttrib;
	DWORD dwFileRecordSize;
	DWORD dwClusterSize;
	DWORD dwBitmapClustersCount;
	
	QWORD qwTotalClustersCount;
	QWORD qwUsedClustersCount;
	QWORD qwFreeClustersCount;
	
	DWORD dwBitmapSize;
	
	char szLabel[128];
	UINT nNtfsVersion;

	unsigned char cReserved[16170]; // Adjust to fit with total header size
};

#endif //CNTFSPART_H
